package app.navigation;

import app.components.label.ValidationLabel;
import app.data.entity.Match;
import app.data.entity.Matchday;
import app.data.entity.Season;
import app.data.service.MatchService;
import app.data.service.MatchdayService;
import app.data.service.SeasonService;
import com.vaadin.flow.component.AbstractField;
import com.vaadin.flow.component.HasValue;
import com.vaadin.flow.component.UI;
import com.vaadin.flow.component.select.Select;
import com.vaadin.flow.router.BeforeEvent;
import com.vaadin.flow.router.HasUrlParameter;
import com.vaadin.flow.router.WildcardParameter;
import org.springframework.beans.factory.annotation.Autowired;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Optional;

@SuppressWarnings("OptionalUsedAsFieldOrParameterType")
public class Navigation implements HasUrlParameter<String> {

    private String route;
    private boolean onlyMatchdaysWithActivity;

    private final List<Runnable> runnablesToBeRunAfterSelection = new ArrayList<>();

    private final SeasonService seasonService;
    private final MatchdayService matchdayService;
    private final MatchService matchService;

    private final List<Season> seasonList = new ArrayList<>();
    private final List<Matchday> matchdayList = new ArrayList<>();
    private final List<Match> matchList = new ArrayList<>();

    private final Select<Season> seasonSelect = new Select<>();
    private final Select<Matchday> matchdaySelect = new Select<>();
    private final Select<Match> matchSelect = new Select<>();

    private boolean seasonEnabled = false;
    private boolean matchdayEnabled = false;
    private boolean matchEnabled = false;

    private final ValidationLabel validationLabel = new ValidationLabel();

    public Navigation(@Autowired SeasonService seasonService,
                      @Autowired MatchdayService matchdayService,
                      @Autowired MatchService matchService) {
        this.seasonService = seasonService;
        this.matchdayService = matchdayService;
        this.matchService = matchService;

        fillSeasonSelectWithData();

        seasonSelect.addValueChangeListener(seasonSelectValueChangeListener());
        matchdaySelect.addValueChangeListener(matchdaySelectValueChangeListener());
        matchSelect.addValueChangeListener(matchSelectValueChangeListener());
    }

    public void setRoute(String route) {
        this.route = route;
    }

    public void setOnlyMatchdaysWithActivity(boolean onlyMatchdaysWithActivity) {
        this.onlyMatchdaysWithActivity = onlyMatchdaysWithActivity;
    }

    public void setSeasonEnabled(boolean seasonEnabled) {
        this.seasonEnabled = seasonEnabled;
    }

    public void setMatchdayEnabled(boolean matchdayEnabled) {
        this.matchdayEnabled = matchdayEnabled;
    }

    public void setMatchEnabled(boolean matchEnabled) {
        this.matchEnabled = matchEnabled;
    }

    public void addRunnableToBeRunAfterSelection(Runnable runnable) {
        runnablesToBeRunAfterSelection.add(runnable);
    }

    public boolean isSeasonEnabled() {
        return seasonEnabled;
    }

    public boolean isMatchdayEnabled() {
        return matchdayEnabled;
    }

    public boolean isMatchEnabled() {
        return matchEnabled;
    }

    private String getRoute() {
        if (route != null) return route;
        throw new IllegalStateException("Route must be set!");
    }

    private void updateUrl() {
        String seasonParam = null;
        String matchdayParam = null;
        String matchParam = null;
        if (seasonEnabled && seasonSelect.getOptionalValue().isPresent()) seasonParam = seasonSelect.getValue().toString();
        if (matchdayEnabled && matchdaySelect.getOptionalValue().isPresent()) matchdayParam = matchdaySelect.getValue().toString();
        if (matchEnabled && matchSelect.getOptionalValue().isPresent()) matchParam = matchSelect.getValue().toString();

        String params = NavigationUtils.getWildcardParam(seasonParam, matchdayParam, matchParam);
        UI.getCurrent().getPage().getHistory().pushState(null, String.format("%s/%s", getRoute(), params));
    }

    private HasValue.ValueChangeListener<? super AbstractField.ComponentValueChangeEvent<Select<Season>, Season>> seasonSelectValueChangeListener() {
        return event -> {
            if (!seasonEnabled) throw new IllegalStateException("Cannot select season when it is not enabled!");
            if (matchdayEnabled) {
                fillMatchdaySelectWithData(event.getValue());
                autoselectMatchday();
                return;
            }
            doPostSelectionStuff();
        };
    }

    private HasValue.ValueChangeListener<? super AbstractField.ComponentValueChangeEvent<Select<Matchday>, Matchday>> matchdaySelectValueChangeListener() {
        return event -> {
            if (!matchdayEnabled) throw new IllegalStateException("Cannot select matchday when it is not enabled!");
            if (matchEnabled) {
                fillMatchSelectWithData(event.getValue());
                autoselectMatch();
                return;
            }
            doPostSelectionStuff();
        };
    }

    private HasValue.ValueChangeListener<? super AbstractField.ComponentValueChangeEvent<Select<Match>, Match>> matchSelectValueChangeListener() {
        return event -> {
            if (!matchEnabled) throw new IllegalStateException("Cannot select match when it is not enabled!");
            doPostSelectionStuff();
        };
    }

    private void doPostSelectionStuff() {
        validationLabel.setValid(true);
        updateUrl();
        runnablesToBeRunAfterSelection.forEach(Runnable::run);
    }

    private void autoselectSeason() {
        if (!seasonEnabled) throw new IllegalStateException("This method should not be called when season is not enabled!");
        if (seasonList.isEmpty()) {
            validationLabel.setText("No Seasons in List!");
            validationLabel.setValid(false);
            return;
        }
        seasonSelect.setValue(seasonList.get(seasonList.size() - 1));
    }

    private void autoselectMatchday() { // TODO: add date stuff and choose depending on date instead!
        if (!matchdayEnabled) throw new IllegalStateException("This method should not be called when matchday is not enabled!");
        if (matchdayList.isEmpty()) {
            validationLabel.setText("No Matchdays in List!");
            validationLabel.setValid(false);
            return;
        }
        Matchday matchdayToSelect = matchdayList.get(0);
        for (Matchday matchday : matchdayList) if (matchdayService.hasActivity(matchday)) matchdayToSelect = matchday;
        matchdaySelect.setValue(matchdayToSelect);
    }

    private void autoselectMatch() {
        if (!matchEnabled) throw new IllegalStateException("This method should not be called when match is not enabled!");
        if (matchList.isEmpty()) {
            validationLabel.setText("No Matches in List!");
            validationLabel.setValid(false);
            return;
        }
        matchSelect.setValue(matchList.get(0));
    }

    private void fillSeasonSelectWithData() {
        seasonList.clear();
        seasonList.addAll(seasonService.getAllSeasonsSorted());
        seasonSelect.setItems(seasonList);
    }

    private void fillMatchdaySelectWithData(Season season) {
        matchdayList.clear();
        List<Matchday> matchdaysToAdd = onlyMatchdaysWithActivity ?
                matchdayService.getMatchdaysWithActivitySorted(season) : matchdayService.getMatchdaysSorted(season);
        matchdayList.addAll(matchdaysToAdd);
        matchdaySelect.setItems(matchdayList);
    }

    private void fillMatchSelectWithData(Matchday matchday) {
        matchList.clear();
        matchList.addAll(matchService.getMatches(matchday));
        matchSelect.setItems(matchList);
    }

    @Override
    public void setParameter(BeforeEvent event, @WildcardParameter String param) {
        Map<UrlParameterType, Optional<String>> map = NavigationUtils.getParameterMap(param);
        navigate(map.get(UrlParameterType.SEASON), map.get(UrlParameterType.MATCHDAY), map.get(UrlParameterType.MATCH));
    }

    private void navigate(Optional<String> seasonParam, Optional<String> matchdayParam, Optional<String> matchParam) {
        if (!seasonEnabled) return;
        Optional<Season> season = getSeasonFromParam(seasonParam);
        if (season.isPresent()) seasonSelect.setValue(season.get());
        else autoselectSeason();

        if (!matchdayEnabled) return;
        Optional<Matchday> matchday = getMatchdayFromParam(matchdayParam);
        if (matchday.isPresent()) matchdaySelect.setValue(matchday.get());
        else autoselectMatchday();

        if (!matchEnabled) return;
        Optional<Match> match = getMatchFromParam(matchParam);
        if (match.isPresent()) matchSelect.setValue(match.get());
        else autoselectMatch();
    }

    private Optional<Season> getSeasonFromParam(Optional<String> seasonParam) {
        if (seasonParam.isEmpty()) {
            return Optional.empty();
        }
        return seasonList.stream()
                .filter(season -> season.toString().equals(seasonParam.get()))
                .findFirst();
    }

    private Optional<Matchday> getMatchdayFromParam(Optional<String> matchdayParam) {
        if (matchdayParam.isEmpty()) {
            return Optional.empty();
        }
        return matchdayList.stream()
                .filter(matchday -> matchday.toString().equals(matchdayParam.get()))
                .findFirst();
    }

    private Optional<Match> getMatchFromParam(Optional<String> matchParam) {
        if (matchParam.isEmpty()) {
            return Optional.empty();
        }
        return matchList.stream()
                .filter(match -> match.toString().equals(matchParam.get()))
                .findFirst();
    }

    public ValidationLabel getValidationLabel() {
        return validationLabel;
    }

    public List<Season> getSeasonList() {
        return seasonList;
    }

    public List<Matchday> getMatchdayList() {
        return matchdayList;
    }

    public List<Match> getMatchList() {
        return matchList;
    }

    public Select<Season> getSeasonSelect() {
        return seasonSelect;
    }

    public Select<Matchday> getMatchdaySelect() {
        return matchdaySelect;
    }

    public Select<Match> getMatchSelect() {
        return matchSelect;
    }

    public Optional<Matchday> getSelectedMatchday() {
        return matchdaySelect.getOptionalValue();
    }

    public Optional<Season> getSelectedSeason() {
        return seasonSelect.getOptionalValue();
    }

    public Optional<Match> getSelectedMatch() {
        return matchSelect.getOptionalValue();
    }

    public SeasonService getSeasonService() {
        return seasonService;
    }

    public MatchdayService getMatchdayService() {
        return matchdayService;
    }

    public MatchService getMatchService() {
        return matchService;
    }
}
